import math
import os
import platform
import re
import shlex
import shutil
import subprocess
import sys
from collections import deque
from io import TextIOWrapper
from pathlib import Path
from typing import List, Dict, Tuple

import click
import rich
from rich.console import Console
from rich.live import Live
from rich.panel import Panel
from schema import SchemaError


class MonitorOutput:
    def __init__(self, title: str, subtitle: str, console: Console, log_path: Path):
        self.title = title
        self.subtitle = subtitle
        self.console = console
        self.log_path = log_path


SENSITIVE_PATTERNS = [
    re.compile(r"API_KEY", re.IGNORECASE),
    re.compile(r"TOKEN", re.IGNORECASE),
    re.compile(r"PASSWORD", re.IGNORECASE),
    re.compile(r"SECRET", re.IGNORECASE),
    re.compile(r"PRIVATE_KEY", re.IGNORECASE),
    re.compile(r"ACCESS_KEY", re.IGNORECASE),
]

HIGH_ENTROPY_THRESHOLD = 3.5


def calculate_entropy(data):
    if not data:
        return 0
    entropy = 0
    for x in set(data):
        p_x = data.count(x) / len(data)
        entropy += -p_x * math.log2(p_x)
    return entropy


def is_potentially_sensitive_key_value(key, value):
    # Check if key matches sensitive patterns
    if any(pattern.search(key) for pattern in SENSITIVE_PATTERNS):
        return True
    # Check if value has high entropy
    if calculate_entropy(value) > HIGH_ENTROPY_THRESHOLD:
        return True
    return False


def validate_str_list(value):
    return isinstance(value, list) and all(isinstance(item, str) for item in value)


def is_yaml_loadable_type(value):
    yaml_loadable_types = (str, int, float, bool, list, dict, type(None))
    if not isinstance(value, yaml_loadable_types):
        raise SchemaError(
            f"Provided type {type(value).__name__} cannot be loaded by yaml.safe_load"
        )
    return value


def bool_to_yn(bool_input: bool) -> str:
    """
    Converts a boolean value to a 'y' or 'n' string representation.

    :param bool_input: The boolean value to be converted.
        - True: Represented by 'y'
        - False: Represented by 'n'
    :return: The string representation of the boolean value.
        - 'y' if bool_input is True
        - 'n' if bool_input is False

    """
    lex = {True: "y", False: "n"}

    return lex[bool_input]


def bool_to_strings(bool_input: bool) -> List[str]:
    """
    Converts a boolean value to a list of corresponding strings.

    :param bool_input: The boolean value to be converted.
    :return: A list of strings representing the boolean value. The list will contain one or more of the following strings:
        - "yes"
        - "y"
        - "true" (if bool_input is True)
        - "no"
        - "n"
        - "false" (if bool_input is False)
    """
    lex = {
        True: ["yes", "y", "true"],
        False: ["no", "n", "false"],
    }

    return lex[bool_input]


def string_to_bool(string_input: str) -> bool or None:
    """
    Convert a string representation of boolean to a boolean value.

    :param string_input: The input string to be converted.
    :type string_input: str
    :return: The corresponding boolean value or None if the input string is not recognized as boolean.
    :rtype: bool or None
    """
    lex = {
        "yes": True,
        "y": True,
        "true": True,
        "no": False,
        "n": False,
        "false": False,
    }

    return lex.get(string_input.lower(), None)


def get_script(script: str, scripts_dict: Dict[str, str], name: str) -> str or None:
    """
    Get the script based on the platform and provided parameters.

    :param script: The name of the script to retrieve.
    :param scripts_dict: A dictionary containing the scripts for different platforms.
    :param name: The name of the tool.
    :return: The script for the given platform and script name, or None if not found.
    :raises ValueError: If the tool does not have the specified script in any platform.
    """
    _os = platform.system().lower()
    windows = _os in ["windows", "win32"]
    macos = _os in ["darwin"]

    _script = None

    # Default to top-level definition of the script (not platform-dependent)
    if script in scripts_dict:
        _script = scripts_dict[script]

    # Any platform-dependent scripts will override the top-level definition
    if type(scripts_dict) is dict:
        if windows and "windows" in scripts_dict.keys():
            if script in scripts_dict["windows"]:
                _script = scripts_dict["windows"][script]
        elif macos and "mac" in scripts_dict.keys():
            if script in scripts_dict["mac"]:
                _script = scripts_dict["mac"][script]
        elif (not windows and not macos) and "linux" in scripts_dict.keys():
            if script in scripts_dict["linux"]:
                _script = scripts_dict["linux"][script]

    if _script is None:
        raise TypeError(
            f"Tool \"{name}\" does not have script '{script}' "
            f"in {list(scripts_dict.keys())}"
        )

    return _script


def type_tool(type_str: str) -> type or None:
    """
    Return the corresponding Python type based on the input string.

    :param type_str: A string representing the desired Python type.
                     Possible values are "int", "float", "str", or "string".
    :return: The corresponding Python type if it exists, otherwise None.
    """
    types = {
        "int": int,
        "float": float,
        "str": str,
        "string": str,
    }
    try:
        return types[type_str.lower()]
    except KeyError:
        return None


def replace_env(string: str) -> str:
    """
    :param string: The string in which to replace environment variable placeholders.
    :return: The string with all environment variable placeholders replaced with their corresponding values.

    This method takes a string as input and replaces all occurrences of environment variable placeholders in the
    format ${variable_name} with their corresponding values. It uses regular expressions to find all placeholders
    in the string, then checks if the corresponding environment variable is set. If the variable is set, it replaces
    the placeholder with the variable's value. If the variable is not set, it does not do any replacement and returns
    the string unchanged.

    Example usage:

    >>> replace_env("Hello ${USERNAME}, your home directory is ${HOME}")
    'Hello John, your home directory is /home/john'
    """
    pattern = re.compile(r"\$\{(.+?)}")
    matches = pattern.findall(string)

    for match in matches:
        env_value = os.getenv(match)
        if env_value is not None:
            string = string.replace(f"${{{match}}}", env_value)

    return string


def run_script_with_env_substitution(
    script_str: str, verbose: bool = False, monitor_output: MonitorOutput or None = None
) -> Tuple[str, int]:
    """
    Run a script with environment variable substitution. If the script fails to run as a shlex'd list, run it as a
    string instead.

    :param monitor_output: Options for running the script with monitored output
    :param verbose: Whether to print warning if the script fails to run as a shlex'd list or not.
    :param script_str: The script to be executed as a string.
    :return: A tuple containing the stdout output and the return code of the script.
    """

    # Inject environment variables
    substituted_script = replace_env(script_str)
    multiline = "\n" in substituted_script

    _script = shlex.split(substituted_script)

    try:
        # If shutil can't find the command or the script is multiline, run as shell
        if shutil.which(_script[0]) is None or multiline:
            if shutil.which(_script[0]) is None and verbose:
                print(
                    f"'{_script[0]}' is not installed. Trying script in shell.",
                    file=sys.stderr,
                )

            _os = platform.system().lower()
            windows = _os in ["windows", "win32"]

            if windows:
                windows_shell = (
                    "pwsh" if shutil.which("pwsh") is not None else "powershell"
                )
                cmd = [windows_shell, "-Command", substituted_script]

                if monitor_output is None:
                    result = subprocess.run(cmd)
                else:
                    result = monitored_subprocess(
                        command=cmd,
                        title=monitor_output.title,
                        subtitle=monitor_output.subtitle,
                        console=monitor_output.console,
                    )
            else:
                if monitor_output is None:
                    result = subprocess.run(
                        substituted_script, shell=True, text=True, capture_output=True
                    )
                else:
                    result = monitored_subprocess(
                        command=substituted_script,
                        title=monitor_output.title,
                        subtitle=monitor_output.subtitle,
                        console=monitor_output.console,
                        shell=True,
                    )
        # Else, run the shlex'd cmd list
        else:
            if monitor_output is None:
                result = subprocess.run(_script, text=True, capture_output=True)
            else:
                result = monitored_subprocess(
                    command=_script,
                    title=monitor_output.title,
                    subtitle=monitor_output.subtitle,
                    console=monitor_output.console,
                )
    except subprocess.CalledProcessError as e:
        return f"{e.stderr.strip()}", e.returncode

    if isinstance(result.stdout, TextIOWrapper):
        result.stdout = result.stdout.read()

    return str(result.stdout), result.returncode


# Code snippet used with permission @Hubro https://github.com/Textualize/rich/discussions/2885#discussioncomment-5382390
def monitored_subprocess(
    command: List[str] or str,
    title: str or None,
    subtitle: str or None,
    console: Console,
    shell: bool = False,
):
    """
    Run a subprocess while displaying the output in a temporary box with Rich
    """
    stdout_output = ""

    box_height = min(max([console.height // 4, 6]), 12)
    box_inner_height = box_height - 2  # The panel has 1 row of padding on each side
    box_width = console.width - 4  # The panel has 2 cols of padding on each side

    buffer = deque(maxlen=box_inner_height)

    def process_panel() -> Panel:
        return Panel(
            "\n".join(buffer), height=box_height, title=title, subtitle=subtitle
        )

    with rich.live.Live(
        get_renderable=process_panel, refresh_per_second=30, transient=True
    ):
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="UTF-8",
            shell=shell,
        )

        assert process.stdout

        try:
            for line in process.stdout:
                buffer.append(line[:box_width].rstrip())
                stdout_output += line
        except KeyboardInterrupt:
            process.terminate()
            raise click.Abort()
        finally:
            # Always wait for the process to terminate, or we might fail later
            # cleanup steps
            process.wait()

        # create a CompletedProcess instance
        completed_process = subprocess.CompletedProcess(
            args=command,
            returncode=process.returncode,
            stdout=stdout_output,
            stderr=process.stderr,
        )

    return completed_process



import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Optional

import typer
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.theme import Theme
from typing_extensions import Annotated

import startout.github_api as gh_api
from startout import util
from startout.env_manager import EnvironmentVariableManager
from startout.init_option import InitOption
from startout.starterfile import parse_starterfile, Starter
from startout.util import replace_env

custom_theme = Theme(
    {
        "input_prompt": "bold cyan",
        "announcement": "bold yellow",
        "success": "bold green",
        "warning": "bold orange1",
        "error": "bold red",
        "bold": "bold",
    }
)

console = Console(theme=custom_theme)
log_path = os.path.join(Path(__file__).parent, "logs", "startout.log")


def prompt_init_option(option: InitOption):
    # If the default type is boolean, then use yes/no parsing
    use_bool = isinstance(option.default, bool)

    if use_bool:
        default_reminder = f" \[default: {util.bool_to_yn(option.default)}]"
    else:
        default_reminder = f" \[default: {option.default}]"

    if use_bool:
        type_reminder = f"(y/n)"
    elif type(option.default) is int:
        type_reminder = f"(enter an integer)"
    elif type(option.default) is float:
        type_reminder = f"(enter a floating point number)"
    else:
        type_reminder = ""

    endl = "\n"
    potential_response = console.input(
        f"[input_prompt]{option.prompt}[/]"
        f"{' ' if len(type_reminder) == 0 else endl}"
        f"[italic cyan]{type_reminder}{default_reminder}: [/]"
    )

    _T = type(option.default)
    response = None
    while response is None:
        if len(potential_response) > 0:
            try:
                if use_bool:
                    response = util.string_to_bool(potential_response)
                else:
                    response = _T(potential_response)
            except ValueError:
                console.print(f"[error]Try again.[/]")
                potential_response = console.input("> ")
        else:
            response = option.default

    return response


# Initialize the typer CLI
startout_paths_app = typer.Typer(name="startout-paths")
startout_starterfile_app = typer.Typer(name="startout-starter")


# fmt: off
@startout_paths_app.command(
    name="init",
    help="Initialize a new Path instance from a template repository and perform all automatic set up steps defined in"
         "its Starterfile"
)
def initialize_path_instance(
        template: Annotated[str, typer.Argument(help="Path Template to use. This may be a fully-formed GitHub repo "
                                                     "(e.g. User/Repo) or part of name which will be completed "
                                                     "interactively")],
        new_repo_name: Annotated[str, typer.Argument(help="Name of new repo created from Path")],
        new_repo_owner: Annotated[Optional[str], typer.Argument(help="User or Organization that will own the new repo "
                                                                     "(leave blank to assign interactively)")] = "",
        public: Annotated[bool, typer.Option("--public/--private", help="Set visibility of the new repo")] = True,
):
    # fmt: on

    #########################
    # Parse the template repo
    #########################

    # Case: User has defined a fully-formed GitHub repo
    if re.match(r"^[^/]*/[^/]*$", template):
        template_owner, template_name = template.split("/", 1)
    # Case: User has defined a non-path, will use interactive mode to form template owner and name
    elif "/" not in template:

        # Interactive mode defaults to using StartOut's Path templates
        result = console.input(
            f"[input_prompt]Please provide the owner of the '{template}' "
            f"Path (case insensitive) \\[default: Start-Out]: [/]",
        )
        if len(result) == 0 or result.lower() == "start-out":

            # StartOut Path templates begin with a prefix 'path-', default behavior is to use this prefix.
            template_owner = "Start-Out"
            template_name = f"path-{template}"
        else:
            template_owner = result
            template_name = template

        console.print(f"INFO: Using template {template_owner}/{template_name}", style='announcement')

    else:
        console.print(
            f"[error]Invalid Path template '{template}', please specify one of the following:[/]\n"
            f"- A fully-formed GitHub repository name (e.g. Owner/Repository)\n"
            f"- A non-path to be defined interactively (e.g. express-react-postgresql)\n"
            f"  * [bold]NOTE[/]: Non-paths will trigger an interactive mode which provides helpful defaults"
        )
        sys.exit(1)

    #######################################
    # Check if defined repository is a Path
    #######################################

    is_path = gh_api.check_repo_custom_property(template_owner, template_name, {
        "Golden-Paths": "Path",
    })

    if not is_path:
        console.print("Warning: Specified template can't be confirmed as a path, some features may be unavailable.",
                      style='warning')

    #################################
    # Parse the owner of the new repo
    #################################

    # If the new owner was not specified when this function was called, the user wants to define it interactively.
    if len(new_repo_owner) == 0:
        new_repo_owner = new_repo_owner_interactive()

    initialized_repo_path = initialize_repo(
        template_owner, template_name, new_repo_owner, new_repo_name, public
    )

    if not initialized_repo_path:
        console.file = sys.stderr  # set console output to stderr
        console.print("Failed to initialize new Path, exiting now.", style='error')
        console.file = sys.stdout  # set console output back to stdout
        sys.exit(1)

    else:
        if not is_path:
            console.print("Not a path: Skipping Starterfile step.", style='warning')
        else:
            # Initialize frameworks using Starterfile
            os.chdir(initialized_repo_path)

            # Start capturing environment variables throughout the process of opening the Path
            env_manager = EnvironmentVariableManager()
            with open("Starterfile.yaml", "r") as starter_file:
                starter = parse_starterfile(starter_file)

            do_starter_init(starter, env_manager)


def do_starter_init(starter: Starter, env_manager: EnvironmentVariableManager):
    init_options = starter.get_init_options()

    responses = {}
    for module_name, options in init_options:
        for option in options:
            response = prompt_init_option(option)
            responses[(module_name, option.name)] = response

    starter.set_init_options(responses)
    starter.up(console, log_path)

    ###############
    # Dump env vars
    if starter.env_dump_file is not None and starter.env_dump_mode is not None:
        # After opening the Path, gather the env vars generated during the process
        console.print(
            f"[info]Collecting environment variables for dump file: {starter.env_dump_file}[/]"
        )
        env_manager.capture_final_env()
        nonsensitive, sensitive = env_manager.get_captured_vars()

        # Ask for approval of each individual entry in sensitive
        approved_sensitive = {}
        for sens_key, sens_val in sensitive.items():
            include_var = console.input(
                f"[input_prompt]Do you want to include the potentially sensitive variable '{sens_key}={sens_val}'? (y/N) [/]")
            if include_var.lower() == 'y':
                approved_sensitive[sens_key] = sens_val

        dump_vars = {**nonsensitive, **approved_sensitive}

        with open(starter.env_dump_file, starter.env_dump_mode) as dump_file:
            if starter.env_dump_mode == 'a':
                dump_file.write("\n")

            for key, val in dump_vars.items():
                dump_file.write(f"{key}={val}\n")

    ########################
    # Do env var replacement
    if starter.env_replacement_targets is not None:
        for target in starter.env_replacement_targets:
            with open(target, 'r') as target_file:
                lines = target_file.readlines()

            with open(target, 'w') as target_file:
                for line in lines:
                    target_file.write(replace_env(line))


def new_repo_owner_interactive() -> str:
    with Progress(SpinnerColumn(), TextColumn("[progress.description]{task.description}")) as progress:

        # Collect valid options for the user using `gh auth status` and `gh org list`

        # Get the current username
        task1 = progress.add_task(f"Checking `gh auth status`", total=None)

        try:
            result = subprocess.run(['gh', 'auth', 'status'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        except FileNotFoundError as e:
            progress.update(task1, description="Failure: gh auth could not be validated", completed=True)
            console.file = sys.stderr  # set console output to stderr
            console.print(f"Failed to run `gh auth status`, make sure `gh` is installed!\n\t{e}", style='error')
            sys.exit(1)

        valid_owners = []

        # Exit if gh auth fails, necessary for the rest of the process
        if result.returncode != 0:
            progress.update(task1, description="Failure: gh auth could not be validated", completed=True)
            console.file = sys.stderr  # set console output to stderr
            if result.stderr is not None:
                console.print(result.stderr.decode(), style='error')
            if result.stdout is not None:
                console.print(result.stdout.decode())
            console.print("Unable to authenticate with GitHub, please ensure you have completed `gh auth login`",
                          style='bold')
            sys.exit(1)

        progress.update(task1, description="Success: gh auth status validated", completed=True)

        # Parse username from gh auth status
        feedback = result.stdout.decode()
        username = re.findall(r"(?<=account )(.*)(?= \(keyring\))", feedback)

        if len(username) != 1:
            console.file = sys.stderr  # set console output to stderr
            console.print("Problem parsing username from `gh auth status`, output was:", style='error')
            console.print(feedback, style='error')
            sys.exit(1)

        # If username is parsed, add to possible owners
        valid_owners.append(username[0])

        # Get authorized orgs via gh org list
        task2 = progress.add_task(f"Checking `gh org list`", total=None)

        result = subprocess.run(['gh', 'org', 'list'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        # Do not exit, but warn the user that this check failed
        if result.returncode != 0:
            console.file = sys.stderr  # set console output to stderr
            console.print(result.stderr.decode(), style='error')
            console.print("Unable to collect valid orgs, please check `gh auth status`", style='error')
            progress.update(task2, description="Failure: Failed to collect orgs", completed=True)
            console.file = sys.stdout  # set console output back to stdout

        else:
            # Parse orgs from command
            feedback = result.stdout.decode()
            lines = feedback.splitlines()

            valid_owners.extend([org for org in lines if len(org) > 0])
            progress.update(task2, description="Success: gh orgs collected", completed=True)

    # All potential new owners are collected, prompt user to choose one
    console.print("Please choose from the following list for the new repo owner:", style='input_prompt')

    i = 0
    for owner in valid_owners:
        console.print(f"[{i}] - [cyan]{owner}[/]")
        i += 1

    choice = None
    flag = False
    while choice is None:
        if flag:
            # Print the invalid feedback every time after the first ask
            console.file = sys.stderr  # set console output to stderr
            console.print("Invalid choice", style='error')
            console.file = sys.stdout  # set console output back to stdout

        flag = True

        choice = console.input("> ")

        try:
            choice = int(choice)
        except ValueError:
            choice = None

        try:
            if choice >= 0:
                return valid_owners[choice]
            else:
                choice = None
        except (IndexError, TypeError):
            choice = None


def initialize_repo(
        template_owner: str or None, template_name: str or None, new_repo_owner: str or None,
        new_repo_name: str or None, public: bool = True
):
    # If any environment variables are missing, prompt the user for them interactively

    # TODO Use some localization scheme for all feedback
    if new_repo_owner is None:
        new_repo_owner = console.input(
            "[input_prompt]Who should be the owner of the new repository?[/]\n"
            "[italic cyan](Must be a user/org authorized with `gh `auth`): [/]"
        )
    if new_repo_name is None:
        new_repo_name = console.input(
            "[input_prompt]What should the name of the new repository be?[/]\n"
            "[italic cyan](Make sure the name isn't already taken): [/]"
        )
    if template_owner is None:
        template_owner = console.input("\n[input_prompt]Who is the owner of the Path template?: [/]")

    if template_name is None:
        template_name = console.input("\n[input_prompt]What is the name of the Path template?: [/]")

    result = gh_api.create_repo_from_temp(
        new_repo_owner, new_repo_name, f"{template_owner}/{template_name}", public
    )

    if not result:
        console.file = sys.stderr  # set console output to stderr
        console.print("Failed to clone Path template.", style='error')
        console.file = sys.stdout  # set console output back to stdout

    else:
        # Update path to the project root if successful
        console.print(f"Cloned new Path to [italic]{result}[/]", style='success')
        os.environ["NEW_PATH_ROOT"] = result

    return result


def startout_paths_command():
    typer.run(initialize_path_instance)


@startout_starterfile_app.command(
    name="up",
    help="Perform all automatic set up steps defined in a Starterfile"
)
def starterfile_up_only(
        starterfile_path: Annotated[Optional[str], typer.Argument(help="Startfile to use")] = "Starterfile.yaml"):
    # Ensure the starterfile path is a valid file
    if not Path(starterfile_path).is_file():
        print(f"No such file '{starterfile_path}'", file=sys.stderr)
        exit(1)

    starterfile_parent_dir = Path(starterfile_path).parent

    # Obtain the current working directory, and then join this path with the 'starterfile_path'
    cwd = os.getcwd()
    full_starterfile_path = os.path.join(cwd, starterfile_path)

    os.chdir(starterfile_parent_dir)

    # Start capturing environment variables throughout the process of opening the Path
    env_manager = EnvironmentVariableManager()
    with open(full_starterfile_path, "r") as starter_file:
        starter = parse_starterfile(starter_file)

    do_starter_init(starter, env_manager)
    os.chdir(cwd)


def startout_starter_app():
    startout_starterfile_app()


if __name__ == "__main__":
    startout_starterfile_app()



import os

from typing import Tuple

from startout.util import is_potentially_sensitive_key_value


class EnvironmentVariableManager:
    def __init__(self):
        self.initial_vars = {k: v for k, v in os.environ.items()}

        self.final_vars = {}

    def capture_final_env(self):
        current_vars = {key: value for key, value in os.environ.items()}
        new_vars = {
            key: value
            for key, value in current_vars.items()
            if key not in self.initial_vars
        }
        self.final_vars.update(new_vars)

        return self.final_vars

    def get_captured_vars(self) -> Tuple[dict, dict]:
        """
        Returns two dictionaries: one containing the not potentially sensitive variables and their values, and the other
        containing the potentially sensitive variables and their values.

        **Returns:**

        - `not_potentially_sensitive` (dict): A dictionary containing the not potentially sensitive variables and their values.
        - `potentially_sensitive` (dict): A dictionary containing the potentially sensitive variables and their values.

        """

        potentially_sensitive = {
            key: value
            for key, value in self.final_vars.items()
            if is_potentially_sensitive_key_value(key, value)
        }
        not_potentially_sensitive = {
            key: value
            for key, value in self.final_vars.items()
            if not is_potentially_sensitive_key_value(key, value)
        }
        return not_potentially_sensitive, potentially_sensitive



from enum import Enum
from typing import List, Dict, Tuple

from schema import Schema, And, Or, Optional

from startout.util import (
    run_script_with_env_substitution,
    get_script,
    validate_str_list,
)


class InstallationMode(Enum):
    INSTALL = "INSTALL"
    OPTIONAL = "OPTIONAL"
    AS_ALT = "AS_ALT"


class InstallationStatus(Enum):
    EXISTING_INSTALLATION = 0
    NEWLY_INSTALLED = 1
    NOT_INSTALLED = 2


def should_rollback(installation_status: InstallationStatus):
    return installation_status == InstallationStatus.NEWLY_INSTALLED


class Tool:
    """
    Class representing a tool with installation and uninstallation scripts.

    Attributes:
        tool_scripts_schema (Schema): Schema definition for tool scripts.
        tool_schema (Schema): Schema definition for the entire tool object.

    Methods:
        __init__(self, name: str, dependencies: list[str], scripts: dict[str, str or dict[str, str]]):
            Initialize a Tool instance with the provided name, dependencies, and scripts.
            Raises a TypeError if 'install' or 'uninstall' scripts are not defined.

        get_script(self, script: str, scripts_list, name: str or None = None) -> str or None:
            Get the script for the given name and platform. May be run basically static, used for instantiation.
            Returns the script string if found, else None.

        run(self, script: str) -> tuple[str, int]:
            Execute the specified script and return the output and return code.
            Returns a tuple of the stdout and return code.

        check(self) -> bool:
            Run the 'check' script and return True if the return code is 0, else False.

        initialize(self) -> bool:
            Run the 'install' script and print the output.
            Returns True if the return code is 0, else False.

        destroy(self) -> bool:
            Run the 'uninstall' script and print the output.
            Returns True if the return code is 0, else False.
    """

    tool_scripts_schema = Schema(
        Or(
            {
                Optional("install"): And(str),
                Optional("uninstall"): And(str),
                Optional("check"): And(str),
                Optional("windows"): {
                    Optional("install"): And(str),
                    Optional("uninstall"): And(str),
                    Optional("check"): And(str),
                },
                Optional("mac"): {
                    Optional("install"): And(str),
                    Optional("uninstall"): And(str),
                    Optional("check"): And(str),
                },
                Optional("linux"): {
                    Optional("install"): And(str),
                    Optional("uninstall"): And(str),
                    Optional("check"): And(str),
                },
            },
        )
    )
    tool_schema = Schema(
        {
            Optional("depends_on"): Or(str, validate_str_list),
            Optional("mode"): Or("install", "optional", "as_alt"),
            Optional("alt"): str,
            "scripts": tool_scripts_schema,
        }
    )

    def __init__(
        self,
        name: str,
        dependencies: List[str] or None,
        scripts: Dict[str, str or Dict[str, str]],
        alt: str or None = None,
        install_mode: str = "INSTALL",
    ):
        """
        Initializes a Tool with the given name, dependencies, and scripts.

        :param name: The name of the tool.
        :param dependencies: A list of dependencies required by the tool.
        :param scripts: A dictionary mapping script names to their respective commands or scripts.

        :raises TypeError: If the 'install' and 'uninstall' scripts are not defined for the module.
        """
        # These calls will raise ValueError if any are missing
        get_script("install", scripts, name=name)
        get_script("uninstall", scripts, name=name)
        get_script("check", scripts, name=name)

        self.name = name
        self.dependencies = dependencies
        self.scripts = scripts
        self.alt = alt

        self.mode = InstallationMode[install_mode.upper()]

        self.status = InstallationStatus.NOT_INSTALLED

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __hash__(self):
        return hash((self.name, str(self.dependencies), str(self.scripts)))

    def run(self, script: str) -> Tuple[str, int]:
        """
        Runs a script with environment variable substitutions.

        :param script: The name of the script to be executed, located in the Tool's scripts.
        :return: A tuple containing the stdout output and the return code of the script execution.
        """
        _script = get_script(script, self.scripts, self.name)

        return run_script_with_env_substitution(_script)

    def check(self):
        """
        Run the Tool's 'check' script.

        :return: True if the response code is 0, False otherwise.
        """
        response, code = self.run("check")

        return code == 0

    def initialize(self):
        """
        Run the Tool's 'install' script.

        :return: True if the response code is 0, False otherwise.
        """
        msg, code = self.run("install")
        print(msg)

        return code == 0

    def destroy(self):
        """
        Run the Tool's 'uninstall' script.

        :return: True if the response code is 0, False otherwise.
        """
        msg, code = self.run("uninstall")
        print(msg)

        return code == 0



import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Dict, Tuple

from rich.console import Console
from schema import Schema, And, Or, Optional, Use

from startout.init_option import InitOption
from startout.util import (
    replace_env,
    run_script_with_env_substitution,
    get_script,
    MonitorOutput,
    monitored_subprocess,
    validate_str_list,
    is_yaml_loadable_type,
)


def check_for_key(name: str, key: str, scripts: dict):
    all_platform_keys = [
        platforms
        for platforms in scripts.keys()
        if platforms in ["windows", "mac", "linux"]
    ]

    # The key must be in the top level (not platform-specific) unless it is specified in EACH platform-specific set

    # If the key is not at the top level...
    if key not in scripts.keys():
        # ...and the three supported platforms are not also all specified...
        if len(all_platform_keys) != 3:
            # ...then it is impossible for the script to have been fully defined.
            raise TypeError(
                f"No '{key}' script defined for module \"{name}\". Failed to create Module."
            )
        else:
            # If all platforms have a set of scripts, make sure that this script is in each of them
            missing_platforms = []
            for _platform in all_platform_keys:
                if key not in scripts[_platform].keys():
                    missing_platforms.append(_platform)
            if len(missing_platforms) > 0:
                raise TypeError(
                    f"Script 'init' not fully defined for module \"{name}\" (missing {missing_platforms}). "
                    f"Failed to create Module."
                )


class Module:
    """
    Class representing a module.

    Module.source defines how the module is collected, defined as:

    - git
      : A URI is passed to `git clone`

    - script
      : A script is executed by `bash` (or `git-bash` on Windows)

    Attributes:
        name (str): The name of the module.
        dest (str): The destination of the module (usually used as the name of the directory into which the module
        is installed).
        source (dict): The source of the module, given as any ONE of [git, curl, script, docker].
        scripts (dict): Scripts associated with the module.
        dependencies (str or list[str]): Dependencies of the module. (Optional)
        init_options (list[dict]): Initialization options for the module. (Optional)

    """

    module_scripts_schema = Schema(
        Or(
            {
                Optional(str): str,
                Optional("windows"): {Optional(str): str},
                Optional("mac"): {Optional(str): str},
                Optional("linux"): {Optional(str): str},
            },
        )
    )
    module_init_options_schema = Schema(
        {
            "env_name": str,
            Optional("type"): is_yaml_loadable_type,
            "default": is_yaml_loadable_type,
            "prompt": str,
        }
    )
    module_schema = Schema(
        {
            "dest": And(str, Use(replace_env)),
            "source": Schema({Or("git", "script", only_one=True): str}),
            "scripts": module_scripts_schema,
            Optional("depends_on"): Or(str, validate_str_list),
            Optional("init_options"): [module_init_options_schema],
        }
    )

    def __init__(
        self,
        name: str,
        dest: str,
        source: str,
        scripts: Dict[str, str],
        dependencies=None,
        init_options=None,
    ):
        """
        Initialize a new Module instance.

        :param name: The name of the module.
        :param dest: The destination path of the module.
        :param source: The source path of the module.
        :param scripts: A dictionary mapping script names to script paths.
        :param dependencies: (optional) A list of module names that this module depends on. Defaults to None.
        :param init_options: (optional) Additional options for module initialization. Defaults to None.
        """

        # These calls will raise an exception if the script is not present
        get_script("init", scripts, name)
        get_script("destroy", scripts, name)

        self.name = name
        self.dest = dest
        self.source = source
        self.scripts = scripts
        self.dependencies = dependencies
        self.init_options = init_options

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(
            (
                self.get_name(),
                self.get_dest(),
                self.get_source(),
                str(self.dependencies),
                str(self.scripts),
                str(self.init_options),
            )
        )

    def get_name(self):
        return replace_env(self.name)

    def get_dest(self):
        return replace_env(self.dest)

    def get_source(self):
        return replace_env(self.source)

    def run(
        self,
        script: str,
        print_output: bool = False,
        monitor_output: MonitorOutput or None = None,
    ) -> Tuple[str, int]:
        """
        Runs a script with environment variable substitutions.

        :param monitor_output:
        :param script: The name of the script to be executed, located in the Module's scripts.
        :param print_output: Whether to print the response at the .... level
        :return: A tuple containing the stdout output and the return code of the script execution.
        """
        if script not in self.scripts:
            raise ValueError(
                f"Module \"{self.name}\" does not have script '{script}' in {list(self.scripts.keys())}"
            )

        response, code = run_script_with_env_substitution(
            get_script(script, self.scripts, self.get_name()),
            monitor_output=monitor_output,
        )

        if print_output and type(response) is str and len(response.strip()) > 0:
            print(f".... [{self.get_name()}.{script}]: {response.strip()}")

        return response, code

    def initialize(
        self, console: Console or None = None, log_path: Path or None = None
    ):
        """
        Run the Module's 'init' script.

        :return: True if the response code is 0, False otherwise.
        """
        mon = None
        if console is not None and log_path is not None:
            mon = MonitorOutput(
                title=f"Initializing {self.get_name()}",
                subtitle="...",
                console=console,
                log_path=log_path,
            )

        msg, code = self.run("init", monitor_output=mon)

        if code != 0:
            print(f".. FAILURE [{self.get_name()}]: {msg}", file=sys.stderr)
            return False
        else:
            print(
                f".. SUCCESS [{self.get_name()}]: Initialized module {self.get_name()}"
            )
            return True

    def destroy(self, console: Console or None = None, log_path: Path or None = None):
        """
        Run the Module's 'destroy' script.

        :return: True if the response code is 0, False otherwise.
        """
        mon = None
        if console is not None and log_path is not None:
            mon = MonitorOutput(
                title=f"Initializing {self.get_name()}",
                subtitle="...",
                console=console,
                log_path=log_path,
            )

        msg, code = self.run("destroy", monitor_output=mon)

        if code != 0:
            print(f".. FAILURE [{self.get_name()}]: {msg}", file=sys.stderr)
            return False
        else:
            print(f".. SUCCESS [{self.get_name()}]: Destroyed module {self.get_name()}")
            return True


class GitModule(Module):
    """
    Module for interacting with Git.

    :class:`GitModule` is a subclass of :class:`Module` and provides functionality
    for initializing a Git repository.

    Example:
        >>> git_module = GitModule()
        >>> git_module.source = "https://github.com/username/repo.git"
        >>> git_module.dest = "/path/to/destination"
        >>> git_module.initialize()

    Note:
        This module requires Git to be installed on the system.

    Attributes:
        source (str): The source URL of the Git repository.
        dest (str): The destination directory to clone the repository into.

    Raises:
        OSError: If Git is not installed on the system.

    """

    def initialize(
        self, console: Console or None = None, log_path: Path or None = None
    ):
        """
        Initializes the object by cloning `self.source` (a repository) to `self.dest` (a directory).

        :return: bool - True if the cloning succeeds, False otherwise.
        :raises OSError: If Git is not installed.
        """
        if shutil.which("git") is None:
            raise OSError(f"Git is not installed. Please install Git and try again.")

        cmd = [
            "git",
            "clone",
            # "--progress",
            self.get_source(),
            self.get_dest(),
        ]

        if console is not None:
            result = monitored_subprocess(
                command=cmd,
                title=f"Cloning {self.get_name()}",
                subtitle="...",
                console=console,
            )
        else:
            result = subprocess.run(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )

        if result.returncode != 0:
            print(
                f".. FAILURE [{self.get_name()}]: {result.stdout.strip()}",
                file=sys.stderr,
            )
            return False
        else:
            print(
                f".... PROGRESS [{self.get_name()}]: Cloned module {self.get_name()}, running init script"
            )

            mon = None
            if console is not None and log_path is not None:
                mon = MonitorOutput(
                    title=f"Initializing {self.get_name()}",
                    subtitle="...",
                    console=console,
                    log_path=log_path,
                )

            msg, code = self.run("init", print_output=True, monitor_output=mon)

            if code != 0:
                print(f".. FAILURE [{self.get_name()}]: {msg}", file=sys.stderr)
                return False
            else:
                print(
                    f".. SUCCESS [{self.get_name()}]: Initialized module {self.get_name()}"
                )
                return True


class ScriptModule(Module):
    def initialize(
        self, console: Console or None = None, log_path: Path or None = None
    ):
        msg, code = run_script_with_env_substitution(self.get_source())

        if code != 0:
            print(f".. FAILURE [{self.get_name()}]: {msg}", file=sys.stderr)
            return False
        else:
            print(
                f".... PROGRESS [{self.get_name()}]: Ran script for module {self.get_name()}"
            )
            if type(msg) is str and len(msg.strip()) > 0:
                print(f".... [{self.get_name()}.source.script]: {msg.strip()}")
            print(f".... PROGRESS [{self.get_name()}]: Running init script")

            mon = None
            if console is not None and log_path is not None:
                mon = MonitorOutput(
                    title=f"Initializing {self.get_name()}",
                    subtitle="...",
                    console=console,
                    log_path=log_path,
                )

            msg, code = self.run("init", print_output=True, monitor_output=mon)

            if code != 0:
                print(f".. FAILURE [{self.get_name()}]: {msg}", file=sys.stderr)
                return False
            else:
                print(
                    f".. SUCCESS [{self.get_name()}]: Initialized module {self.get_name()}"
                )
                return True


def create_module(module: dict, name: str):
    """
    Create a module object based on the given parameters. The value of module.source determines the type of module.

    :param module: A dictionary representing the module information.
    :param name: The name of the module.
    :return: An instance of a module object.
    """
    mode = next(iter(module["source"]))
    source = module["source"][mode]
    dest = module["dest"]

    options = None
    if "init_options" in module.keys():
        options = []
        for options_set in module["init_options"]:
            options.append(InitOption(options_set))
    dependencies = None
    if "depends_on" in module.keys():
        _deps = module["depends_on"]
        if type(_deps) is str:
            dependencies = [_deps]
        else:
            dependencies = _deps

    # Instantiate the correct type of Module
    _T = Module
    if mode == "git":
        _T = GitModule
    elif mode == "script":
        _T = ScriptModule

    return _T(
        name=name,
        dest=dest,
        source=source,
        scripts=module["scripts"],
        init_options=options,
        dependencies=dependencies,
    )



from startout.util import replace_env, type_tool


class InitOption:
    """
    Initializes an instance of the `InitOption` class. This class is used as an API for setting options before a Module
    is initialized (e.g. name of the project, passwords, etc.).

    This API is meant to be used by tools which interface with the Starterfile parser, e.g. the Startout CLI.

    :param options_set: A dictionary containing the options for initialization.
                        The dictionary should have the following keys:
                        - "default" (required): The default value for the option.
                        - "type" (optional): The type of the option value.
                        - "env_name" (required): The name of the environment variable associated with the option.
                        - "prompt" (required): The prompt to display when prompting for the option value.
    """

    def __init__(self, options_set):
        """
        Initializes the method.

        :param options_set: A dictionary containing the options for the method.
                    - "default" (required): The default value for the option. This value may not be `None`
                    - "type" (optional): The type of the option value.
                    - "env_name" (required): The name of the environment variable associated with the option.
                    - "prompt" (required): The prompt to display when prompting for the option value.
        """
        default = options_set["default"]

        assert default is not None

        if "type" in options_set.keys():
            _t = type_tool(options_set["type"])
            default = _t(default)

        self.name = replace_env(options_set["env_name"])
        self.default = replace_env(default) if type(default) is str else default
        self.prompt = replace_env(options_set["prompt"])
        self.value = None



import json
import os.path
import shlex
import subprocess
import sys
from rich.console import Console
from rich.theme import Theme
from rich.progress import Progress, SpinnerColumn, TextColumn

custom_theme = Theme(
    {
        "input_prompt": "bold cyan",
        "announcement": "bold yellow",
        "success": "bold green",
        "error": "bold red",
        "bold": "bold",
    }
)

console = Console(theme=custom_theme)


def create_repo_from_temp(
    owner: str, repo_name: str, template: str, public: bool = False
):
    """
    Create a GitHub repo from a provided template.

    :param owner: The owner of the new repo to be created, must be a GitHub user/org for which the currently
    authenticated (`gh auth status`) has permission to create a repo.
    :param repo_name: The name of the new repo to be created.
    :param template: The owner and name of the template repo e.g. "StartOut/template-repo"
    :param public: The visibility for the repo to be created (defaults to private)
    """

    cmd_string = f"gh repo create {owner}/{repo_name} --template={template} --clone"
    cmd = shlex.split(cmd_string)

    if public:
        cmd.append("--public")
    else:
        cmd.append("--private")

    result = ""

    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}")
    ) as progress:
        task = progress.add_task(f"Fetching and cloning {template}", total=None)

        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        if result.returncode == 0:
            progress.update(
                task,
                description=f"Success: {template} has been cloned.",
                completed=True,
            )
            progress.refresh()
            progress.stop()
            console.print(f"{result.stdout.decode()}", style="success")
            return os.path.join(os.getcwd(), repo_name)
        else:
            progress.update(
                task,
                description=f"Failure: {template} has not been cloned.",
                completed=True,
            )
            progress.refresh()
            progress.stop()
            console.file = sys.stderr  # Set console output to stderr
            console.print(f"ERROR: {result.stdout.decode()}", style="error")
            return False


def check_repo_custom_property(
    template_owner: str, template_name: str, custom_properties: dict
) -> bool:
    cmd_string = f"gh api /repos/{template_owner}/{template_name}/properties/values"
    cmd = shlex.split(cmd_string)

    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}")
    ) as progress:
        task = progress.add_task(f"Checking properties of {template_name}", total=None)

        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        if result.returncode == 0:
            # Try to get a dict from result
            response_str = result.stdout.decode()

            repo_properties = {}
            try:
                for _property in json.loads(response_str):
                    repo_properties[_property["property_name"]] = _property["value"]

            except ValueError:
                progress.update(
                    task,
                    description=f"Failure: Failed to check properties of {template_name} "
                    f"(response {response_str}).",
                    completed=True,
                )
                progress.refresh()
                progress.stop()
                return False

            progress.update(
                task,
                description=f"Success: Received custom properties of {template_name}.",
                completed=True,
            )
            progress.refresh()

            if custom_properties != repo_properties:
                progress.update(task, description=f"Expected: {custom_properties}")
                progress.update(task, description=f"Received: {repo_properties}")
                progress.update(
                    task,
                    description=f"Failure: Mismatch in properties of {template_name}.",
                    completed=True,
                )
                progress.refresh()
                progress.stop()
                return False

            progress.stop()
            console.print(
                f"Validated all custom properties for {template_name}", style="success"
            )
            return True
        else:
            progress.update(
                task,
                description=f"Failure: Could not check properties of {template_name}.",
                completed=True,
            )
            progress.refresh()
            progress.stop()
            console.file = sys.stderr  # Set console output to stderr
            console.print(f"ERROR: {result.stdout.decode()}", style="error")
            return False



import itertools
import os
import sys
from pathlib import Path
from typing import TextIO, List, Tuple

import yaml
from dotenv import load_dotenv
from rich.console import Console
from schema import Schema, And, Or, Optional, Use

from startout.module import Module, create_module
from startout.tool import Tool, InstallationStatus, should_rollback, InstallationMode
from startout.util import replace_env


class Starter:
    """
    Starter class

    The Starter class is used to install modules and tools required for a project. It allows for easy management of
    module and tool dependencies.

    Use:
        1. use parse_starterfile on a valid Starterfile.yaml to generate a Starter
        2. use Starter.get_init_options() to collect any init options that the Starterfile may have
        3. use Starter.set_init_options() to set any init options
        4. use Starter.up() to perform the initialization of the project

    Attributes:
        starterfile_schema (Schema): Schema definition for the starter file.

    Methods:
        __init__(modules, tools, module_dependencies, tool_dependencies)
            Initializes a new instance of the Starter class.

            Parameters:
                modules (list[Module]): List of modules to be installed.
                tools (list[Tool]): List of tools to be installed.
                module_dependencies (list[list[str]]): List of module dependencies in layers.
                tool_dependencies (list[list[str]]): List of tool dependencies in layers.

        up(teardown_on_failure=True)
            Installs the modules and tools.

            Parameters:
                teardown_on_failure (bool, optional): Specifies whether to rollback successfully installed tools or
                 modules if any others fail.
                Defaults to True.

            Returns:
                bool: True if installation is successful, False otherwise.

        install_tools(teardown_on_failure=True)
            Installs the tools. Called by up().

            Parameters:
                teardown_on_failure (bool, optional): Specifies whether to rollback installations on failure.
                Defaults to True.

            Returns:
                bool: True if installation is successful, False otherwise.

        install_modules(teardown_on_failure=True)
            Installs the modules. Called by up().

            Parameters:
                teardown_on_failure (bool, optional): Specifies whether to rollback installations on failure.
                Defaults to True.

            Returns:
                bool: True if installation is successful, False otherwise.

        get_init_options()
            Placeholder method for getting the initialization options.

        set_init_options(options)
            Placeholder method for setting the initialization options.
    """

    env_dump_schema = Schema(
        {
            "target": And(str, len),
            Optional("mode"): Or("a", "w", "A", "W"),
        }
    )
    starterfile_schema = Schema(
        {
            "tools": And(dict, len),
            "modules": And(dict, len),
            Optional("env_file"): And(Or(Use(list), None)),
            Optional("env_replace"): And(list, len),
            Optional("env_dump"): env_dump_schema,
        }
    )

    def __init__(
        self,
        modules: List[Module],
        tools: List[Tool],
        module_dependencies: List[List[str]],
        tool_dependencies: List[List[str]],
        env_replacement_targets: List[str] = None,
        env_dump: Tuple[str, str] = None,
    ):
        """
        Initializes the class instance with given modules, tools, module dependencies, and tool dependencies.

        :param modules: A list of Module objects representing the parsed modules.
        :param tools: A list of Tool objects representing the parsed tools.
        :param module_dependencies: A list of lists of strings representing the dependencies between modules.
            Each inner list depends on one or more of the modules in the previous inner list (the first list has no
            dependencies).
        :param tool_dependencies: A list of lists of strings representing the dependencies between tools.
            Each inner list depends on one or more of the modules in the previous inner list (the first list has no
            dependencies).
        """
        self.modules = modules
        self.tools = tools
        self.module_dependencies = module_dependencies
        self.tool_dependencies = tool_dependencies
        self.env_replacement_targets = env_replacement_targets

        if env_dump is None:
            self.env_dump_file = None
            self.env_dump_mode = None
        else:
            self.env_dump_file = env_dump[0]
            self.env_dump_mode = env_dump[1]

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            modules_match = self.modules == other.modules
            tools_match = self.tools == other.tools
            module_deps_match = self.module_dependencies == other.module_dependencies
            tool_deps_match = self.tool_dependencies == other.tool_dependencies

            return (
                modules_match and tools_match and module_deps_match and tool_deps_match
            )

    def up(
        self,
        console: Console or None = None,
        log: Path or None = None,
        teardown_on_failure=True,
        fail_early=True,
    ):
        """
        Installs all Tools, all Modules, and performs environment variable replacement on a Startersteps.md file (if
        applicable)

        :param console:
        :param log:
        :param teardown_on_failure: A boolean flag to determine whether to perform teardown operations if any failure occurs during the method execution. Default value is `True`.
        :param fail_early: A boolean flag to determine whether to abort the process as soon as a tool or module fails to initialize. Default value is `False`.
        :return: A boolean value indicating whether the tools and modules installation was successful. Returns `True` if both tools and modules were installed successfully, otherwise returns `False`.
        """
        tools_installed = self.install_tools(
            teardown_on_failure, fail_early, console, log
        )
        modules_installed = self.install_modules(
            console, log, teardown_on_failure, fail_early
        )

        if not tools_installed:
            print("ERROR: Failed to install tools!", file=sys.stderr)
        if not modules_installed:
            print("ERROR: Failed to install modules!", file=sys.stderr)

        # If a Startersteps.md file is present, perform environment variable replacement
        if Path("Startersteps.md").is_file():
            new_lines = []
            with open("Startersteps.md", "r") as steps_in:
                new_lines.extend(steps_in.readlines())

            with open("Startersteps.md", "w") as steps_out:
                for line in new_lines:
                    steps_out.write(replace_env(line))

        return tools_installed and modules_installed

    def install_tools(
        self,
        teardown_on_failure=True,
        fail_early=True,
        console: Console or None = None,
        log: Path or None = None,
        assumption: bool or None = None,
    ):
        """
        Install tools layer by layer so that their dependencies are all met before being installed.

        :param assumption: When asking for input, assumes True or False if set
        :param teardown_on_failure: If True, rollback other tools if any tool installation fails.
        :type teardown_on_failure: bool
        :param fail_early: If True, function will return false as soon as a tool fails to initialize.
        :type fail_early: bool
        :return: True if all tools are installed successfully, False otherwise.
        :rtype: bool
        """

        # Early exit if there are no tools to install.
        if len(self.tools) == 0:
            print("Nothing to do.")
            return False

        print("Installing tools...")

        for tool in (
            tool for tool in self.tools if tool.mode == InstallationMode.OPTIONAL
        ):
            if assumption is None:
                potential_response = (
                    console.input(
                        f"[input_prompt]Install {tool.name}? (y/N): [/]"
                    ).lower
                    == "y"
                )
            else:
                potential_response = assumption

            tool.mode = (
                InstallationMode.INSTALL
                if potential_response
                else InstallationMode.OPTIONAL
            )

        failed_tools = []
        successful_tools = []
        current_layer = -1
        for layer in self.tool_dependencies:
            # Install tools layer by layer so that their dependencies are all met before being installed
            current_layer += 1
            early_exit = False

            for tool in (
                tool
                for tool in self.tools
                if tool.name in layer
                if tool.mode == InstallationMode.INSTALL
            ):
                # Use the tool's check function to prevent attempts to install an existing tool
                if tool.check():
                    print(f".. Tool '{tool.name}' is already installed, skipping.")
                    tool.status = InstallationStatus.EXISTING_INSTALLATION
                    continue

                # Initialize this tool, adding it to the list of failures if it cannot be initialized
                if not tool.initialize():
                    if tool.alt is None:
                        print(f".. Tool '{tool.name}' failed to install.")
                        failed_tools.append(tool.name)
                        tool.status = InstallationStatus.NOT_INSTALLED
                        if fail_early:
                            early_exit = True
                    else:
                        alt = next((t for t in self.tools if t.name == tool.alt), None)

                        print(
                            f".. Tool '{tool.name}' failed to install, will use alt '{alt.name}' instead."
                        )
                        alt.mode = InstallationMode.INSTALL

                        # Add the alt to the next dependency layer if it is not already accounted for
                        if alt.name not in [
                            tool_name
                            for layer in self.tool_dependencies[current_layer:]
                            for tool_name in layer
                        ]:
                            self.tool_dependencies[current_layer].append(alt.name)

                else:
                    print(f".. Tool '{tool.name}' installed successfully.")
                    successful_tools.append(tool.name)
                    tool.status = InstallationStatus.NEWLY_INSTALLED

            if early_exit:
                break

        # Upon any failed tools, report which tools failed...
        if len(failed_tools) > 0:
            print("Failed tools:", failed_tools, file=sys.stderr)

            # ... and teardown if specified
            if teardown_on_failure:
                tools_to_rollback = [
                    tool
                    for tool in self.tools
                    if tool.name in successful_tools and should_rollback(tool.status)
                ]

                print(
                    "Rolling back these installed tools:",
                    tools_to_rollback,
                    file=sys.stderr,
                )

                destroyed_tools = []
                for tool in tools_to_rollback:
                    if not tool.destroy():
                        # TODO handle failure to destroy better
                        print(
                            f'FATAL: Failed to destroy tool "{tool.name}"',
                            file=sys.stderr,
                        )
                        print(
                            f".. Only destroyed these tools: {destroyed_tools}",
                            file=sys.stderr,
                        )
                        sys.exit(1)
                    else:
                        destroyed_tools.append(tool.name)

            # If any tools failed
            return False

        # If all tools were successfully installed or were already installed
        return True

    def install_modules(
        self,
        console: Console or None = None,
        log: Path or None = None,
        teardown_on_failure=True,
        fail_early=True,
    ):
        """
        Install modules layer by layer so that their dependencies are all met before being installed.

        :param log:
        :param console:
        :param teardown_on_failure: If True, rollback other tools if any module installation fails.
        :type teardown_on_failure: bool
        :param fail_early: If True, function will return false as soon as a tool fails to initialize.
        :type fail_early: bool
        :return: True if all modules are installed successfully, False otherwise.
        :rtype: bool
        """

        # Early exit if there are no modules to install.
        if len(self.modules) == 0:
            print("Nothing to do.")
            return False

        print("Installing modules...")

        failed_modules = []
        successful_modules = []
        for layer in self.module_dependencies:
            # Install modules layer by layer so that their dependencies are all met before being installed
            early_exit = False

            for module in (module for module in self.modules if module.name in layer):
                # Initialize this module, adding it to the list of failures if it cannot be initialized
                if not module.initialize(console=console, log_path=log):
                    failed_modules.append(module.name)
                    if fail_early:
                        early_exit = True
                else:
                    successful_modules.append(module.name)

            if early_exit:
                break

        # Report any failed modules
        if len(failed_modules) > 0:
            print("Failed modules:", failed_modules, file=sys.stderr)

            # Destroy modules which were successfully installed if specified
            if teardown_on_failure:
                print(
                    "Rolling back successfully installed modules:",
                    successful_modules,
                    file=sys.stderr,
                )

                destroyed_modules = []
                for module in [
                    module
                    for module in self.modules
                    if module.name in successful_modules
                ]:
                    if not module.destroy(console=console, log_path=log):
                        # TODO handle failure to destroy better
                        print(
                            f'FATAL: Failed to destroy module "{module.name}"',
                            file=sys.stderr,
                        )
                        print(
                            f".. Only destroyed these modules: {destroyed_modules}",
                            file=sys.stderr,
                        )
                        sys.exit(1)
                    else:
                        destroyed_modules.append(module.name)

            # If any modules failed to initialize
            return False

        # If all modules were successfully initialized
        return True

    def get_init_options(self):
        """
        Returns a list of tuples containing the name and init_options of modules
        that have a non-null value for init_options in the self.modules list.

        :return: A list of tuples where each tuple contains the name and init_options
                 of modules satisfying the condition.
        """
        return [
            (module.name, module.init_options)
            for module in self.modules
            if module.init_options is not None
        ]

    def set_init_options(self, options):
        """
        Sets the init options of the Starter to be used before calling .up()

        :param options: A dictionary of module and option names along with their corresponding values.
                        Example: {("react", "MODULE_REACT_USE_TYPESCRIPT"): True,
                                  ("react", "MODULE_REACT_APP_NAME"): "example-react-app"}
        :return: None
        """
        for module_name, option_name in options:
            # Get the value from the options set
            value = options[(module_name, option_name)]

            # Update the environment
            os.environ[option_name] = str(value)

            # Update the internal variable of the option itself
            module = [module for module in self.modules if module.name == module_name][
                0
            ]
            option = [
                option for option in module.init_options if option.name == option_name
            ][0]
            option.value = value


def create_dependency_layers(items: List[Module or Tool]) -> List[List[str]]:
    """
    Generate dependency layers based on the given items such that each inner list's dependencies are all contained
    within the preceding inner list (the first inner list has no dependencies).

    :param items: a list of `Module` or `Tool` objects representing the items
    :type items: list[Module or Tool]
    :return: a list of lists, each containing item names grouped by their dependencies
    :rtype: list[list[str]]
    """
    # Handle dependencies
    dependency_layers = [[item.name for item in items if item.dependencies is None]]

    # Check for any unfulfilled dependencies,
    #  collect all items that are in any item's depends_on fields
    all_items_depended_upon = set(
        itertools.chain.from_iterable(
            [item.dependencies for item in items if item.dependencies is not None]
        )
    )

    # Check if any items that are depended upon are not present (e.g. 'a' requires 'z' but only 'a' and 'b' are defined)
    unmet_dependencies = all_items_depended_upon - set([item.name for item in items])

    if len(unmet_dependencies) > 0:
        print(f"ERROR: Dependency not met {unmet_dependencies}", file=sys.stderr)
        sys.exit(1)

    # Modules with no dependencies are added to the first layer

    # Put remaining modules (those with dependencies) in a set
    dependent_items = False
    if len(dependency_layers) > 0:
        dependent_items = set(
            [item for item in items if item.name not in dependency_layers[0]]
        )

    # Add items to layers until none are left
    while dependent_items:
        scheduled_items = list(itertools.chain.from_iterable(dependency_layers))

        added_items = []
        for dependent_item in dependent_items:
            if all(needed in scheduled_items for needed in dependent_item.dependencies):
                added_items.append(dependent_item)

        if len(added_items) == 0:
            print(
                f"ERROR: Could not meet dependencies for {[item.name for item in dependent_items]}, "
                f"may be a circular dependency.",
                file=sys.stderr,
            )
            sys.exit(1)

        for added_item in added_items:
            dependent_items.remove(added_item)

        dependency_layers.append([item.name for item in added_items])

    return dependency_layers


def parse_starterfile(starterfile_stream: TextIO) -> Starter:
    """
    Parse a Starterfile.yaml

    Checks the Starterfile, Modules, Tools, and their constituent dependencies against their respective schemas and
    creates a `Starter` complete with dependency ordering for tools and modules.

    :param starterfile_stream: The stream representing the starter file.
    :type starterfile_stream: TextIO
    :return: The parsed Starter object.
    :rtype: Starter
    """
    loaded = yaml.safe_load(starterfile_stream)

    if "env_file" in loaded.keys():
        if type(loaded["env_file"]) is list:
            for env_file in loaded["env_file"]:
                _path = os.path.join(os.path.dirname(starterfile_stream.name), env_file)
                load_dotenv(str(_path))
        elif type(loaded["env_file"]) is str:
            _path = os.path.join(
                os.path.dirname(starterfile_stream.name), loaded["env_file"]
            )
            load_dotenv(str(_path))

    Starter.starterfile_schema.validate(loaded)

    tools = []

    for tool_name in loaded["tools"]:
        _tool = loaded["tools"][tool_name]
        tool = Tool.tool_schema.validate(_tool)

        dependencies = tool["depends_on"] if "depends_on" in _tool else None
        mode = tool["mode"] if "mode" in _tool else "INSTALL"
        alt = tool["alt"] if "alt" in _tool else None

        if type(dependencies) is str:
            dependencies = [dependencies]

        if alt is not None:
            dependencies.append(alt)

        tools.append(Tool(tool_name, dependencies, tool["scripts"], alt, mode))

    tool_dependencies = create_dependency_layers(tools)

    print("SUCCESS! Parsed tools:", [tool.name for tool in tools])

    modules = []

    for module_name in loaded["modules"]:
        _module = loaded["modules"][module_name]
        module = Module.module_schema.validate(_module)

        modules.append(create_module(module, module_name))

    module_dependencies = create_dependency_layers(modules)

    print("SUCCESS! Parsed modules:", [module.get_name() for module in modules])

    env_replacement_targets = (
        loaded["env_replace"] if "env_replace" in loaded.keys() else None
    )
    if "env_dump" in loaded.keys():
        env_dump_file, env_dump_mode = (
            loaded["env_dump"]["target"],
            loaded["env_dump"]["mode"],
        )
        env_dump = (env_dump_file, env_dump_mode)
    else:
        env_dump = None

    return Starter(
        modules=modules,
        tools=tools,
        module_dependencies=module_dependencies,
        tool_dependencies=tool_dependencies,
        env_replacement_targets=env_replacement_targets,
        env_dump=env_dump,
    )
